def hello(str_name="gangster"):
    '''
    This is a sample function that returns a string.
    
    Parameters
    ----------
    str_name : str, optional
        [description], by default "gangster"
    
    Returns
    -------
    str
        The string returned.
    '''
    return "hi %s!" % (str_name) 